#include "AGILe/Sherpa/Sherpa.hh"
#include <iostream>
#include <sstream>
#include <cstring>
#include <unistd.h>
#include <sys/stat.h>

#include "Sherpa.H"
#include "Input_Output_Handler.H"
#include "HepMC2_Interface.H"
#include "Event_Handler.H"
#include "Data_Reader.H"
#include "Run_Parameter.H"
#include "Initialization_Handler.H"
#include "Read_Write_Base.H"
#include "Random.H"
using namespace SHERPA;
using namespace ATOOLS;


/// Implementation of Sherpa generator interface.
/// @author Andy Ilott
/// @author Andy Buckley
namespace AGILe {

  // Standard constructor
  Sherpa::Sherpa() {
    // Set up particle names map
    _particleNames[ELECTRON]    =    "11";
    _particleNames[POSITRON]    =   "-11";
    _particleNames[PROTON]      =  "2212";
    _particleNames[ANTIPROTON]  = "-2212";
    //   _particleNames[NEUTRON]     = "N";
    //   _particleNames[ANTINEUTRON] = "NBAR    ";
    //   _particleNames[PHOTON]      = "GAMMA   ";
    //   _particleNames[MUON]        = "MU-     ";
    //   _particleNames[ANTIMUON]    = "MU+     ";
    //   // TAU, ANTITAU
    //   _particleNames[NU_E]        = "NU_E    ";
    //   _particleNames[NU_EBAR]     = "NU_EBAR ";
    //   _particleNames[NU_MU]       = "NU_MU   ";
    //   _particleNames[NU_MUBAR]    = "NU_MUBAR";
    //   _particleNames[NU_TAU]      = "NU_TAU  ";
    //   _particleNames[NU_TAUBAR]   = "NU_TAUBR";
    //   _particleNames[PIPLUS]      = "PI+     ";
    //   _particleNames[PIMINUS]     = "PI-     ";
    //   // PIZERO
    //   // PHOTOELECTRON, PHOTOPOSITRON,
    //   // PHOTOMUON,     PHOTOANTIMUON,
    //   // PHOTOTAU,      PHOTOANTITAU,
  }


  void Sherpa::setSeed(const int value) {
    ATOOLS::ran.SetSeed(value);
  }


  // Set up initial state from supplied parameters
  void Sherpa::setGenSpecificInitialState(PdgCode p1, double e1, 
                                            PdgCode p2, double e2) {
    getLog() << Log::INFO << "Setting initial state..." << endl;
    setParam("BEAM_ENERGY_1", e1);    setParam("BEAM_ENERGY_2", e2);
    setParam("BEAM_1", _particleNames[p1]);
    setParam("BEAM_2", _particleNames[p2]);
    //_initialstateset = true;
  }


  void _makeDirIfNeeded(const string& dir) {
    struct stat buf;
    if (stat(dir.c_str(), &buf) != 0) {
      system(("mkdir " + dir).c_str());
    }
    if (access(dir.c_str(), R_OK & W_OK & X_OK) != 0) {
      throw runtime_error("Can't write to directory: " + dir);
    }
  }


  // Run generator initialization
  void Sherpa::initialize() {
    Generator::initialize();
    getLog() << Log::INFO << "Initialising..." << endl; 

    // // Tell Sherpa where to load its integration results.
    // if (_parameters.find("RESULT_DIRECTORY") == _parameters.end()) {
    //   _parameters["RESULT_DIRECTORY"] = "AGILeSherpaIntResults"; 
    // }
    // // Make dir if necessary.
    // _makeDirIfNeeded(_parameters["RESULT_DIRECTORY"]);

    // // Tell Sherpa where to make/load its ME libraries.
    // if (_parameters.find("SHERPA_CPP_PATH") == _parameters.end()) {
    //   // Note: this path must be absolute.
    //   /// @todo Think about placing these somewhere specific to the model being used, cf. hash codes.
    //   // Relevant files which would require a clean or a different hash code are:
    //   //  {Processes,Particle,ME,Integration,Selector}.dat
    //   _parameters["SHERPA_CPP_PATH"] = string(getenv("PWD")) + "/AGILeSherpaProcessLibs";
    // }
    // // Make dir if necessary.
    // _makeDirIfNeeded(_parameters["SHERPA_CPP_PATH"]);

    // // Specify the path which Sherpa uses to find configuration files.
    // /// @todo Intention is that SHERPA will use PATH to look for a directory
    // /// relative to PWD, and fall back if needed to SHERPADATADIR/Run/$PATH
    // if (_parameters.find("PATH") == _parameters.end()) {
    //   getLog() << Log::ERROR << "Currently need to specify a PATH param to SHERPA" << endl;
    //   throw runtime_error("Currently need to specify a PATH param to SHERPA");
    //   /// @todo Get this working without a prefix path: Sherpa change.
    //   //_parameters["PATH"] = "LHC";
    //   //_parameters["PATH"] = string(SHERPADATADIR) + "/Run/" + LEP91;
    // }


    // Convert the parameter map into a vector which can be passed to Sherpa.
    /// @todo Get a better interface function put in Sherpa so we don't have to do this crap!!!
    vector<char*> c;
    c.push_back("DummyZerothParam");
    for (map<string,string>::const_iterator it = _parameters.begin(); it != _parameters.end(); ++it) {
      string command = (it->first + "=" + it->second);
      char* part = (char*) malloc((command.length()+1)*sizeof(char));
      c.push_back(strcpy(part, command.c_str()));
    }
    try {
      getLog() << Log::INFO << "Initialising run:" << endl;
      _sherpa.InitializeTheRun(c.size(), &c.front());
    } catch (ATOOLS::Exception& e) {
      cerr << e.Info() << endl;
      if (e.Type() == ATOOLS::ex::normal_exit) {
        cerr << "Normal exit..." << endl;
        exit(1);
      }
      //exit(1);
      getLog() << Log::ERROR << "Unexpected error during Sherpa run initialisation..." << endl;
      throw(e);
    }
    getLog() << Log::INFO << "Initialising event handler:" << endl;
    _sherpa.InitializeTheEventHandler(); 

    _initialized = true;
  }


  /// Run the generator for one event
  void Sherpa::makeEvent(HepMC::GenEvent& evt) {
    Generator::makeEvent(evt);
    _sherpa.GenerateOneEvent(); 
    fillEvent(evt);
  }


  /// Fill a HepMC event
  void Sherpa::fillEvent(HepMC::GenEvent& evt) {
    _hepmcInterface.Sherpa2HepMC(_sherpa.GetEventHandler()->GetBlobs(), evt);
  }


  /// Tidy up, print out run stats, etc.
  void Sherpa::finalize() {
    getLog() << Log::INFO << "Finalising..." << endl;
    _sherpa.SummarizeRun(); 
  }


}


// Class factory
extern "C" {
  AGILe::Generator* create() { return new AGILe::Sherpa(); }
  void destroy(AGILe::Generator* gen) { delete gen; }
}
